#!/bin/bash

#set -e -o pipefail
shopt -s failglob

die() {
  echo "$@" 1>&2
  exit 1
}

check_cmd() {
  which $1 >/dev/null 2>&1 || die "$0: Unable to find '$1' in PATH"
}

# help
if [ "$1" = "-h" ] || [ "$1" = "--help" ];then
  echo "ebr-ltl-synth.sh filename [-smv|-tlsf] [-s|--synth]"
  exit 0
fi

# parameters
if [ -z "$1" ];then
  echo "*** Please specify the input file."
  exit 1
fi

synth=""
if [ "$3" = "-s" ] || [ "$3" = "--synth" ];then
  synth="true"
else
  synth="false"
fi


### Global definitions

# Test availability of tools
check_cmd syfco
check_cmd ltl2smv
check_cmd nuXmv
check_cmd demiurge

# result of a safety call  
safetyres="false"
# controllable and uncontrollable variables of a (generic) LTL synthesis problem
controllables="" 
uncontrollables=""
# number of total variables
total_vars="0"
# startTime and endTime
startTime=""
endTime=""
# timeout
timeout="timeout"
timeoutres=""

### FUNCTIONS
starttime()
{
  startTime=$(python -c "import time; print(int(time.time()*1000))")
}

endtime()
{
  endTime=$(python -c "import time; print(int(time.time()*1000))")
}

get_controllables()
{
  filename=$1
  controllables=$(syfco -outs $filename)
}

get_uncontrollables()
{
  filename=$1
  uncontrollables=$(syfco -ins $filename)
}


# From TLSF to SMV
TLSF_to_SMV()
{
  filename=$1
  outfile=$2
  # from TLSF to SMV (+NNF).
  syfco -f smv -o "${outfile}.smv" $filename
  # retaining only φ in "${filename}.smv": this is done because "ltl2smv" tool
  # takes only a formula in input, without keywords like MODULE, VAR , LTLSPEC etc...
  grep -e 'LTLSPEC' "${outfile}.smv" | sed 's/LTLSPEC//g' > "${outfile}.smv1"
  mv "${outfile}.smv1" "${outfile}.smv"
}



# From 𝜑 to Symbolic Deterministic Safety Automaton A(𝜑) 
to_automaton()
{
  tlsfname=$1
  
  # get controllables
  get_controllables $tlsfname
  # get uncontrollables
  get_uncontrollables $tlsfname
  
  filename=$2

  # removing "INPUT" and "OUTPUT" from $filename
  formula="$filename.smv"

  # transforming φ into Symbolic Deterministic Safety Automaton A(𝜑)
  ltl2smv -d 0 ${formula} &> ${filename}.automaton.smv


  # building controllable and uncontrollable parameters
  params="($controllables"
  # decides if adding the comma or not
  if [ -z "$uncontrollables" ];then
    params="$params )"
  elif [ -z "$controllables" ];then
    params="$params $uncontrollables)"
  else
    params="$params , $uncontrollables)"
  fi
  
  # adding the parameters at the end of the line of
  # "${filename}.automaton.smv" containing MODULE definition
  sed -i -e "/^MODULE/ s/$/$params/" ${filename}.automaton.smv

  # cleaning
  rm -f $formula
  rm -f ${filename}.automaton.smv-e
}


# Adding the MAIN module
add_main()
{
  filename=$1
  maindef="
\n-- Main
MODULE main
IVAR\n"

  # adding the definitions of controllables and uncontrollables variables  
  IFS=', ' read -r -a arrayContr <<< "$controllables"
  IFS=', ' read -r -a arrayUncontr <<< "$uncontrollables"
  for var in "${arrayContr[@]}"
  do
    maindef="$maindef\t$var : boolean;\n"
    # increment total_vars
    total_vars=$(expr $total_vars + 1)
  done
  for var in "${arrayUncontr[@]}"
  do
    maindef="$maindef\t$var : boolean;\n"
    # increment total_vars
    total_vars=$(expr $total_vars + 1)
  done

  # creating a new instance of A(φ)
  maindef="${maindef}VAR\n\tphi : ltl_spec_0("
  # adding controllables and uncontrollables variables as (comma-separated) 
  # parameters
  maindef="$maindef $controllables"
  # decides if adding the comma or not
  if [ -z "$uncontrollables" ];then
    maindef="$maindef"
  elif [ -z "$controllables" ];then 
    maindef="$maindef $uncontrollables"
  else
    maindef="$maindef , $uncontrollables"
  fi
  maindef="$maindef );\n"
  # appending the MAIN module to the automaton file
  printf "$maindef" >> $filename.automaton.smv
}


# transformation to AIGER format
to_AIGER()
{
  filename=$1
  echo "set input_file $filename.automaton.smv; go; build_boolean_model; fsmv2aig -f \"$filename.automaton.aag\" -c \"$controllables\"; quit;" > ${filename}.nuxmv-commands.txt
  # writes the functional model into $filename.automaton.aag; log into "nuxmv.$k.output"
  nuXmv -source ${filename}.nuxmv-commands.txt &> ${filename}.nuxmv.output
  
  rm ${filename}.nuxmv-commands.txt
  rm ${filename}.nuxmv.output
}

# call to Demiurge tool
callDemiurge()
{
  [ -z "$DEMIURGETP" ] && die "Please set the DEMIURGETP variable"
  filename=$1
  if [ "$synth" = "false" ];then
    demiurge -r --print="EW" -i "$filename.automaton.aag" &> "$filename.demiurge.result"
  else
    demiurge --print="EWR" -i "$filename.automaton.aag" &> "$filename.demiurge.result"
  fi
}

# Main cycle of the algorithm.
execute()
{
  filename=$1
  safetyres="false"
 
  callDemiurge $filename

  # for synthesis
  if [ $synth = "true" ];then
    sed -i 's/\[RES\] The specification is realizable./REALIZABLE/g' "$filename.demiurge.result"
    sed -i 's/\[RES\] The specification is unrealizable./UNREALIZABLE/g' "$filename.demiurge.result"
  fi
  
  # prints the result
  cat "$filename.demiurge.result"
  
  # cleans the auxiliary files
  rm -f $filename.demiurge.result
  rm -f $filename.automaton.smv
  rm -f $filename.automaton.aag
}



### MAIN

# from TLSF to SMV
name=$(basename -s .tlsf $1)
outfile="$2/$name"
TLSF_to_SMV $1 ${outfile}
# from 𝜑 to Symbolic Deterministic Safety Automaton A(𝜑) 
to_automaton $1 ${outfile}
# from A(𝜑) to AIGER
add_main ${outfile}
to_AIGER ${outfile}
# call to safety synthesizer
execute ${outfile}
