#!/bin/bash

#set -e -o pipefail
shopt -s failglob

die() {
  echo "$@" 1>&2
  exit 1
}

check_cmd() {
  which $1 >/dev/null 2>&1 || die "$0: Unable to find '$1' in PATH"
}

# help
if [ "$1" = "-h" ] || [ "$1" = "--help" ];then
  echo "ebr-ltl-synth.sh filename [-smv|-tlsf] -b [demiurge | safetysynth] [-s|--synth] "
  exit 0
fi

# parameters
if [ -z "$1" ];then
  echo "*** Please specify the input file."
  exit 1
fi

format=""
if [ "$2" == "-tlsf" ];then 
  format="tlsf"
elif [ "$2" == "-smv" ];then 
  format="smv"
else
  echo "*** The second parameter can be either -smv (default) or -tlsf."
  exit 1
fi

backend=""
if [ "$4" = "demiurge" ];then
  backend="demiurge" 
elif [ "$4" = "safetysynth" ];then
  backend="safetysynth"	
else
  echo "Unknown backend."
  exit 1 
fi

synth=""
if [ "$5" = "-s" ] || [ "$5" = "--synth" ];then
  synth="true"
else
  synth="false"
fi



### Global definitions

# Test availability of tools
check_cmd syfco
check_cmd ltl2smv
check_cmd nuXmv
check_cmd demiurge
check_cmd SafetySynth

# result of a safety call  
safetyres="false"
# controllable and uncontrollable variables of a (generic) LTL synthesis problem
controllables="" 
uncontrollables=""
# number of total variables
total_vars="0"
# startTime and endTime
startTime=""
endTime=""
# timeout
timeout="timeout"
timeoutres=""

### FUNCTIONS
starttime()
{
  startTime=$(python -c "import time; print(int(time.time()*1000))")
}

endtime()
{
  endTime=$(python -c "import time; print(int(time.time()*1000))")
}

get_controllables()
{
  name=$1
  if [ $format == "tlsf" ];then
    controllables=$(syfco -outs $name.tlsf)
  else
    controllables=$(awk '{if($1 ~ /OUTPUT/){{for (i=2; i<=NF; i++) if($i != ":") printf $i;} printf "\n"}}' "${name}.smv")
  fi

}

get_uncontrollables()
{
  name=$1
  if [ $format == "tlsf" ];then
    uncontrollables=$(syfco -ins $name.tlsf)
  else
    uncontrollables=$(awk '{if($1 ~ /INPUT/){{for (i=2; i<=NF; i++) if($i != ":") printf $i;} printf "\n"}}' "${name}.smv")
  fi
}


# From TLSF to SMV
TLSF_to_SMV()
{
  name=$1
  outfile=$2
  # from TLSF to SMV (+NNF).
  syfco -f smv -o "${outfile}.smv" $name
  # retaining only φ in "${name}.smv": this is done because "ltl2smv" tool
  # takes only a formula in input, without keywords like MODULE, VAR , LTLSPEC etc...
  grep -e 'LTLSPEC' "${outfile}.smv" | sed 's/LTLSPEC//g' > "${outfile}.smv1"
  mv "${outfile}.smv1" "${outfile}.smv"
}



# From 𝜑 to Symbolic Deterministic Safety Automaton A(𝜑) 
to_automaton()
{
  filename=$1
  
  # get controllables
  get_controllables $filename
  # get uncontrollables
  get_uncontrollables $filename

  # removing "INPUT" and "OUTPUT" from $filename
  formula=""
  if [ $format == "smv" ];then
    sed '/INPUT/d' "${filename}.smv" > "${filename}.smv1"
    sed '/OUTPUT/d' "${filename}.smv1" > "${filename}.smv2"
    formula="${filename}.smv2"
    rm "${filename}.smv1"
  else
    formula="$filename.smv"
  fi

  # transforming φ into Symbolic Deterministic Safety Automaton A(𝜑)
  ltl2smv -d 0 ${formula} &> ${filename}.automaton.smv

  # building controllable and uncontrollable parameters
  params="($controllables"
  # decides if adding the comma or not
  if [ -z "$uncontrollables" ];then
    params="$params )"
  elif [ -z "$controllables" ];then
    params="$params $uncontrollables)"
  else
    params="$params , $uncontrollables)"
  fi
  
  # adding the parameters at the end of the line of
  # "${filename}.automaton.smv" containing MODULE definition
  sed -i -e "/^MODULE/ s/$/$params/" ${filename}.automaton.smv

  # cleaning
  rm -f $formula
}


# Adding the MAIN module
add_main()
{
  name=$1
  maindef="
\n-- Main
MODULE main
IVAR\n"

  # adding the definitions of controllables and uncontrollables variables  
  IFS=', ' read -r -a arrayContr <<< "$controllables"
  IFS=', ' read -r -a arrayUncontr <<< "$uncontrollables"
  for var in "${arrayContr[@]}"
  do
    maindef="$maindef\t$var : boolean;\n"
    # increment total_vars
    total_vars=$(expr $total_vars + 1)
  done
  for var in "${arrayUncontr[@]}"
  do
    maindef="$maindef\t$var : boolean;\n"
    # increment total_vars
    total_vars=$(expr $total_vars + 1)
  done

  # creating a new instance of A(φ)
  maindef="${maindef}VAR\n\tphi : ltl_spec_0("
  # adding controllables and uncontrollables variables as (comma-separated) 
  # parameters
  maindef="$maindef $controllables"
  # decides if adding the comma or not
  if [ -z "$uncontrollables" ];then
    maindef="$maindef"
  elif [ -z "$controllables" ];then 
    maindef="$maindef $uncontrollables"
  else
    maindef="$maindef , $uncontrollables"
  fi
  maindef="$maindef );\n"
  # appending the MAIN module to the automaton file
  printf "$maindef" >> $name.automaton.smv
}


# transformation to AIGER format
to_AIGER()
{
  name=$1
  echo "set input_file $name.automaton.smv; go; build_boolean_model; fsmv2aig -f \"$name.automaton.aag\" -c \"$controllables\"; quit;" > ${name}.nuxmv-commands.txt
  # writes the functional model into $name.automaton.aag; log into "nuxmv.$k.output"
  nuXmv -source ${name}.nuxmv-commands.txt &> ${name}.nuxmv.output
  rm ${name}.nuxmv-commands.txt
  rm ${name}.nuxmv.output
}

# call to Demiurge tool
callDemiurge()
{
  [ -z "$DEMIURGETP" ] && die "Please set the DEMIURGETP variable"
  name=$1
  if [ "$synth" = "false" ];then
    if [ "$backend" = "demiurge" ];then
      demiurge -r --print="EW" -i "$name.automaton.aag" &> "$name.$backend.result"
    else
      SafetySynth "$name.automaton.aag" &> "$name.$backend.result"
    fi
  else
    if [ "$backend" = "demiurge" ];then
      demiurge --print="EWR" -i "$name.automaton.aag" &> "$name.$backend.result"
    else
      SafetySynth --synthesize "$name.automaton.aag" &> "$name.$backend.result"
    fi
  fi
}

# Main cycle of the algorithm.
execute()
{
  name=$1
  safetyres="false"
 
  callDemiurge $name

  # for synthesis
  if [ "$backend" = "demiurge" ];then
    sed -i 's/\[RES\] The specification is realizable./REALIZABLE/g' "$name.$backend.result"
    sed -i 's/\[RES\] The specification is unrealizable./UNREALIZABLE/g' "$name.$backend.result"
  else
    sed -i 's/^realizable/REALIZABLE/g' "$name.$backend.result"
    sed -i 's/^unrealizable/UNREALIZABLE/g' "$name.$backend.result"
  fi
  
  # prints the result
  cat "$name.$backend.result"
  
  # cleans the auxiliary files
  rm -f $name.$backend.result
  rm -f $name.automaton.smv
  rm -f $name.automaton.aag
}



### MAIN

# from TLSF to SMV
name=$(basename -s .tlsf $(basename -s .smv $1))
path=$(dirname $1)
filename=$path/$name
if [ $format == "tlsf" ];then 
  TLSF_to_SMV $1 ${filename}
fi
# from 𝜑 to Symbolic Deterministic Safety Automaton A(𝜑) 
to_automaton ${filename} 
# from A(𝜑) to AIGER
add_main ${filename}
to_AIGER ${filename}
# call to safety synthesizer
execute ${filename}
